#!/usr/bin/env python
'''FFS Recovery Tool

This provides a set of classes and a command line utility to access
UFS1 and UFS2 file systems indepentant of the OS.  It provides a way
to directly read files by inode number along with dumping all directory
inodes even if they are not currently linked into the filesystem
hierarchy.

If the freebsd python extensions are installed, it supports restoring
flags, otherwise it will print out flags it is unable to restore.

WARNING: If snapshots are supported on the fs, dummy superblocks may
be found.'''

__author__ = 'John-Mark Gurney <jmg@FreeBSD.org>'
__copyright__ = 'Copyright 2005 John-Mark Gurney'
__license__ = 'BSD'
__version__ = 'alpha'

__all__ = [
	'DEV_BSIZE', 'daddr_size',			# Data
	'FFSCylinderGroup', 'FFSSuperBlock', 'FFS',	# Classes
	'getsbblk', 'getsboff',				# Functions
	'getoffsetof',
	'findsuperblocks', 'findstartfs', 'printsb',
	]

import block
import dpkt
import getopt
import itertools
import os
import re
import struct
import sys
import weakref
from itertools import ifilter, imap
try:
	import freebsd
	def doflags(name, flags, desc = ''):
		'''Restore flags.'''
		freebsd.lchflags(name, flags)
except ImportError:
	def doflags(name, flags, desc):
		'''Warn if flags are to be restored.'''
		if flags:
			print 'Cannot restore flags: %s, to %s.' % (desc, os.path.join(os.getcwd(), name))

#If there is no l*, alias them to just *
def aliasfun(arg):
	n, m = arg
	if n not in dir(m):
		setattr(m, n, getattr(m, n[1:]))

filter(aliasfun, (
			('lchown', os),
			('lchmod', os),
		))

ptr = {4: 'i', 8: 'q' }[struct.calcsize('P')]
DEV_BSIZE = 512
DIRBLKSIZ = DEV_BSIZE
daddr_size = { 'UFS1': 4, 'UFS2': 8 }
daddr_struct = { 'UFS1': 'i', 'UFS2': 'q' }
NXADDR = 2		# External addresses in inode
NDADDR = 12		# Direct addresses in inode
NIADDR = 3		# Indirect addresses in inode
endianness = '<'	# Little endian for now

class FFSCylinderGroup(dpkt.Packet):
	"""FFS Cylinder Group"""

	CG_MAGIC = 0x090255

	def isvalid(self):
		'''Check to make sure that we are valid.  Currently only
checkes the magic.'''

		if self.cg_magic == self.CG_MAGIC:
			return True
		else:
			return False

	__hdr__ = (
		('cg_firstfield', 'i', -1),
		('cg_magic', 'i', -1),
		('cg_old_time', 'i', -1),
		('cg_cgx', 'i', -1),
		('cg_old_ncyl', 'h', -1),
		('cg_old_niblk', 'h', -1),
		('cg_ndblk', 'i', -1),
		('cg_cs_ndir', 'i', -1),
		('cg_cs_nbfree', 'i', -1),
		('cg_cs_nifree', 'i', -1),
		('cg_cs_nffree', 'i', -1),
		('cg_rotor', 'i', -1),
		('cg_frotor', 'i', -1),
		('cg_irotor', 'i', -1),
		('cg_frsum', '8i', (-1, -1, -1, -1, -1, -1, -1, -1)),
		('cg_old_btotoff', 'i', -1),
		('cg_old_boff', 'i', -1),
		('cg_iusedoff', 'i', -1),
		('cg_freeoff', 'i', -1),
		('cg_nextfreeoff', 'i', -1),
		('cg_clustersumoff', 'i', -1),
		('cg_clusteroff', 'i', -1),
		('cg_nclusterblks', 'i', -1),
		('cg_niblk', 'i', -1),
		('cg_initediblk', 'i', -1),
		('cg_sparecon32', '3i', (-1, -1, -1)),
		('cg_time', 'q', -1),
		('cg_sparecon64', '3q', (-1, -1, -1)),
		('cg_space', '1B', (0)),
		)
	__byte_order__ = endianness

def getsbblk(f, blk):
	'''Return a FFSSuperBlock instance in f starting at blk.'''

	return FFSSuperBlock(f.getlen(blk, FFSSuperBlock.SBLOCK_SIZE))

def getsboff(f, off):
	'''Like getsbblk, but takes a byte offset instead of a blk number.'''

	assert off % f.blocksize == 0
	return getsbblk(f, off / f.blocksize)

class FFSSuperBlock(dpkt.Packet):
	"""FFS Super Block"""

	FS_MAGIC = { 0x011954: 'UFS1', 0x19540119: 'UFS2', 0x19960408: 'BAD' }

	FS_UFS1_MAGIC = 0x011954
	FS_UFS2_MAGIC = 0x19540119
	FS_BSD_MAGIC = 0x19960408

	SBLOCK_SIZE = 8192

	def isvalid(self):
		'''Checks to make sure the superblock has a valid magic, and
that fields that can be calculated from each other are equal.'''

		if self.fs_magic in self.FS_MAGIC and self.check():
			return True
		else:
			return False

	def check(self):
		'''Validate that fields that can be computed from each other
are valid.'''

		tests = (
			lambda x: x.fs_bmask == 0xffffffffl & ~(x.fs_bsize - 1),
			lambda x: x.fs_fmask == 0xffffffffl & ~(x.fs_fsize - 1),
			lambda x: (1 << x.fs_bshift) == x.fs_bsize,
			lambda x: (1 << x.fs_fshift) == x.fs_fsize,
			lambda x: x.fs_fragshift == x.fs_bshift - x.fs_fshift,
			# this will have to change if we support other sector
			# sizes
			lambda x: (DEV_BSIZE << x.fs_fsbtodb) == x.fs_fsize,
			lambda x: x.fs_qbmask == x.fs_bsize - 1,
			lambda x: x.fs_qfmask == x.fs_fsize - 1,
			)
		typespec = {
			'UFS1': (
			lambda x: x.fs_old_nspf == x.fs_fsize / DEV_BSIZE,
			lambda x: x.fs_nindir == x.fs_bsize / daddr_size['UFS1'],
			#lambda x: x.fs_inopb == x.fs_bsize / ,
			),
			'UFS2': (
			lambda x: x.fs_nindir == x.fs_bsize / daddr_size['UFS2'],
			#lambda x: x.fs_inopb == x.fs_bsize / ,
			),
		}
		for v, i in enumerate(tests + typespec[self.getmagic()]):
			if not i(self):
				raise ValueError, 'Test #%d is invalid' % v

		return True

	def getmagic(self):
		'''Return the FS type.  Is one of 'UFS1', 'UFS2', 'BAD' (for
incompletely newfs'd fs's) and 'invalid'.'''
		try:
			return self.FS_MAGIC[self.fs_magic]
		except KeyError:
			return 'invalid'

	blkoff = lambda self, loc: loc & self.fs_qbmask
	fragoff = lambda self, loc: loc & self.fs_qfmask
	lfragtosize = lambda self, frag: frag << self.fs_fshift
	lblktosize = lambda self, size: blk << self.fs_bshift
	smallblktosize = lambda self, blk: blk << self.fs_bshift
	lblkno = lambda self, loc: loc >> self.fs_bshift
	numfrags = lambda self, loc: loc >> self.fs_fshift
	blkroundup = lambda self, size: (size + self.fs_qbmask) & self.fs_bmask
	fragroundup = lambda self, size: (size + self.fs_qfmask) & self.fs_fmask
	fragstoblks = lambda self, frags: frags >> self.fs_fragshift
	blkstofrags = lambda self, blks: blks << self.fs_fragshift
	fragnum = lambda self, fsb: fsb & (self.fs_frag - 1)
	blknum = lambda self, fsb: fsb & ~(self.fs_frag - 1)

	cgbase = lambda self, c: self.fs_fpg * c
	cgdmin = lambda self, c: self.cgstart(c) + self.fs_dblkno
	cgimin = lambda self, c: self.cgstart(c) + self.fs_iblkno
	cgsblock = lambda self, c: self.cgstart(c) + self.fs_sblkno
	cgtod = lambda self, c: self.cgstart(c) + self.fs_cblkno
	def cgstart(self, c):
		if self.fs_magic == self.FS_UFS2_MAGIC:
			return self.cgbase(c)
		return self.cgbase(c) + self.fs_old_cgoffset * (c & ~self.fs_old_cgmask)

	fsbtodb = lambda self, b: b << self.fs_fsbtodb
	dbtofsb = lambda self, b: b >> self.fs_fsbtodb

	ino_to_cg = lambda self, x: x / self.fs_ipg
	ino_to_fsba = lambda self, x: self.cgimin(self.ino_to_cg(x)) +	\
			self.blkstofrags((x % self.fs_ipg) / self.fs_inopb)
	ino_to_fsbo = lambda self, x: x % self.fs_inopb

	dtog = lambda self, d: d / self.fs_fpg
	dtogd = lambda self, d: d % self.fs_fpg

	__hdr__ = (
		('fs_firstfield', 'i', -1),
		('fs_unused_1', 'i', -1),
		('fs_sblkno', 'i', -1),
		('fs_cblkno', 'i', -1),
		('fs_iblkno', 'i', -1),
		('fs_dblkno', 'i', -1),
		('fs_old_cgoffset', 'i', -1),
		('fs_old_cgmask', 'i', -1),
		('fs_old_time', 'i', -1),
		('fs_old_size', 'i', -1),
		('fs_old_dsize', 'i', -1),
		('fs_ncg', 'i', -1),
		('fs_bsize', 'i', -1),
		('fs_fsize', 'i', -1),
		('fs_frag', 'i', -1),
		('fs_minfree', 'i', -1),
		('fs_old_rotdelay', 'i', -1),
		('fs_old_rps', 'i', -1),
		('fs_bmask', 'I', -1),
		('fs_fmask', 'I', -1),
		('fs_bshift', 'i', -1),
		('fs_fshift', 'i', -1),
		('fs_maxcontig', 'i', -1),
		('fs_maxbpg', 'i', -1),
		('fs_fragshift', 'i', -1),
		('fs_fsbtodb', 'i', -1),
		('fs_sbsize', 'i', -1),
		('fs_spare1', '2i', (-1, -1)),
		('fs_nindir', 'i', -1),
		('fs_inopb', 'i', -1),
		('fs_old_nspf', 'i', -1),
		('fs_optim', 'i', -1),
		('fs_old_npsect', 'i', -1),
		('fs_old_interleave', 'i', -1),
		('fs_old_trackskew', 'i', -1),
		('fs_id', '2i', (-1, -1)),
		('fs_old_csaddr', 'i', -1),
		('fs_cssize', 'i', -1),
		('fs_cgsize', 'i', -1),
		('fs_spare2', 'i', -1),
		('fs_old_nsect', 'i', -1),
		('fs_old_spc', 'i', -1),
		('fs_old_ncyl', 'i', -1),
		('fs_old_cpg', 'i', -1),
		('fs_ipg', 'i', -1),
		('fs_fpg', 'i', -1),
		('fs_old_cstotal_cs_ndir', 'i', -1),
		('fs_old_cstotal_cs_nbfree', 'i', -1),
		('fs_old_cstotal_cs_nifree', 'i', -1),
		('fs_old_cstotal_cs_nffree', 'i', -1),
		('fs_fmod', 'b', -1),
		('fs_clean', 'b', -1),
		('fs_ronly', 'b', -1),
		('fs_old_flags', 'b', -1),
		('fs_fsmnt', '468s', ""),
		('fs_volname', '32s', ""),
		('fs_swuid', 'Q', 0),
		('fs_pad', 'i', -1),
		('fs_cgrotor', 'i', -1),
		('fs_ocsp', '%d%s' % (128 / struct.calcsize("P") - 4, ptr),
			(0,) * (128 / struct.calcsize("P") - 4)),
		('fs_contigdirs', ptr, 0),
		('fs_csp', ptr, 0),
		('fs_maxcluster', 'i', -1),
		('fs_active', ptr, 0),
		('fs_old_cpc', 'i', -1),
		('fs_maxbsize', 'i', -1),
		('fs_sparecon64', '17q', (-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1)),
		('fs_sblockloc', 'q', -1),
		('fs_cstotal_cs_ndir', 'q', -1),
		('fs_cstotal_cs_nbfree', 'q', -1),
		('fs_cstotal_cs_nifree', 'q', -1),
		('fs_cstotal_cs_nffree', 'q', -1),
		('fs_cstotal_cs_numclusters', 'q', -1),
		('fs_cstotal_cs_spare', '3q', (-1, -1, -1)),
		('fs_time', 'q', -1),
		('fs_size', 'q', -1),
		('fs_dsize', 'q', -1),
		('fs_csaddr', 'q', -1),
		('fs_pendingblocks', 'q', -1),
		('fs_pendinginodes', 'i', -1),
		('fs_snapinum', '20i', (-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1)),
		('fs_avgfilesize', 'i', -1),
		('fs_avgfpdir', 'i', -1),
		('fs_save_cgsize', 'i', -1),
		('fs_sparecon32', '26i', (-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1)),
		('fs_flags', 'i', -1),
		('fs_contigsumsize', 'i', -1),
		('fs_maxsymlinklen', 'i', -1),
		('fs_old_inodefmt', 'i', -1),
		('fs_maxfilesize', 'Q', 0),
		('fs_qbmask', 'q', -1),
		('fs_qfmask', 'q', -1),
		('fs_state', 'i', -1),
		('fs_old_postblformat', 'i', -1),
		('fs_old_nrpos', 'i', -1),
		('fs_spare5', '2i', (-1, -1)),
		('fs_magic', 'i', -1),
		)
	__byte_order__ = endianness

class Direct(dpkt.Packet):
	'''An entry in a directory.  Pass in the directory string.'''

	__hdr__ = (
		('d_ino', 'I', 0),
		('d_reclen', 'H', 0),
		('d_type', 'B', 0),
		('d_namlen', 'B', 0),
		)
	__byte_order__ = endianness

	ino = property(lambda x: x.d_ino)
	type = property(lambda x: x.d_type)
	name = property(lambda x: x.data[:x.d_namlen])

	def __repr__(self):
		magicshiftnumber = 12	# shift d_type up to Inode mode
		return '<Direct: ino: %d, reclen: %d, type: %s, name: %s>' % (self.ino, self.d_reclen, InodeConsts().ifmttostr(self.type << magicshiftnumber), repr(self.name))

class DirBlock:
	'''Pass in an entire block from a directory.  It will automaticly
subdivide the block into DIRBLKSIZ parts, and iterate over each
entry in the directory.'''

	def __init__(self, data):
		self.data = data

	def __iter__(self):
		i = self.data
		l = 0
		for d in [i[x * DIRBLKSIZ: (x + 1) * DIRBLKSIZ] for x in range(len(i) / DIRBLKSIZ)]:
			while d:
				entry = Direct(d)
				if entry.d_reclen == 0 or entry.d_reclen > DIRBLKSIZ or entry.d_namlen == 0:
					print 'invalid record:', repr(entry)
					break
				yield entry
				d = d[entry.d_reclen:]
			l += 1

class InodeConsts:
	'''Provide various constants used for inodes.'''

	IEXEC =  0000001	# Executable
	IWRITE = 0000002	# Writeable
	IREAD =  0000004	# Readable
	ISVTX =  0001000	# Sticky bit
	ISGID =  0002000	# Set-gid
	ISUID =  0004000	# Set-uid

	IFMT  =  0170000	# Mask of file type
	IFIFO =  0010000	# Named pipe (fifo)
	IFCHR =  0020000	# Character device
	IFDIR =  0040000	# Directory file
	IFBLK =  0060000	# Block device
	IFREG =  0100000	# Regular file
	IFLNK =  0120000	# Symbolic link
	IFSOCK = 0140000	# UNIX domain socket
	IFWHT =  0160000	# Whiteout
	IFMT_STR = {
		IFIFO: 'FIFO',
		IFCHR: 'CHR',
		IFDIR: 'DIR',
		IFBLK: 'BLK',
		IFREG: 'REG',
		IFLNK: 'LNK',
		IFSOCK: 'SOCK',
		IFWHT: 'WHT',
		}
	IFMT_SSTR = {
		IFIFO: 'p',
		IFCHR: 'c',
		IFDIR: 'd',
		IFBLK: 'b',
		IFREG: '-',
		IFLNK: 'l',
		IFSOCK: 's',
		IFWHT: 'w',
		}

	UF_NODUMP	= 0x00000001
	UF_IMMUTABLE	= 0x00000002
	UF_APPEND	= 0x00000004
	UF_OPAQUE	= 0x00000008
	UF_NOUNLINK	= 0x00000010
	SF_ARCHIVED	= 0x00010000
	SF_IMMUTABLE	= 0x00020000
	SF_APPEND	= 0x00040000
	SF_NOUNLINK	= 0x00100000
	SF_SNAPSHOT	= 0x00200000
	FLAGS = {
		UF_NODUMP:	'nodump',
		UF_IMMUTABLE:	'nouchg',
		UF_APPEND:	'uappnd',
		UF_OPAQUE:	'opaque',
		UF_NOUNLINK:	'uunlnk',
		SF_ARCHIVED:	'arch',
		SF_IMMUTABLE:	'schg',
		SF_APPEND:	'sappnd',
		SF_NOUNLINK:	'sunlnk',
		SF_SNAPSHOT:	'snapshot',
		}

	def ifmttostr(self, ifmt):
		return self.IFMT_STR.get(ifmt, 'INVLD')

	def ifmttosstr(self, ifmt):
		return self.IFMT_SSTR.get(ifmt, 'X')

	def getflags(self):
		'''Returns short forms of the flags.  If there are no flags set
it returns `-'.'''

		ret = []
		f = self.di_flags
		while f:
			t = f & -f
			try:
				ret.append(self.FLAGS[t])
			except KeyError:
				ret.append('unknown%d' % t)
			f &= ~t

		if not ret:
			return '-'

		return ','.join(ret)

	def ifmt(self):
		'''Return the masked value of IFMT.'''
		return self.di_mode & self.IFMT

	def getifmt(self):
		'''Returns a string representing the type of this Inode.'''
		return self.ifmttosstr(self.ifmt())

	def getmode(self):
		'''Returns just the mode part.  It maasks off the ifmt.'''
		return self.di_mode & ~self.IFMT

	def getusermode(self):
		'''The user modes shifted for use with IREAD, IWRITE and IEXEC.'''
		return (self.di_mode >> 6) & 0x7

	def getgroupmode(self):
		'''The group modes shifted for use with IREAD, IWRITE and IEXEC.'''
		return (self.di_mode >> 3) & 0x7

	def getothermode(self):
		'''The other modes shifted for use with IREAD, IWRITE and IEXEC.'''
		return self.di_mode & 0x7

	def isdir(self):
		'''Is this inode a directory?'''
		return (self.di_mode & self.IFMT) == self.IFDIR

	def isreg(self):
		'''Is this inode a regular file?'''
		return (self.di_mode & self.IFMT) == self.IFREG

	def islnk(self):
		'''Is this inode a symlink?'''
		return (self.di_mode & self.IFMT) == self.IFLNK

	def lsfmt(self):
		'''The permissions/type of the inode in ls -l style format.'''

		mode = self.di_mode
		ret = [ '%c' % self.getifmt() ]
		m = self.getusermode()
		if m & self.IREAD:
			ret.append('r')
		else:
			ret.append('-')
		if m & self.IWRITE:
			ret.append('w')
		else:
			ret.append('-')
		if mode & self.ISUID and not m & self.IEXEC:
			ret.append('S')
		elif mode & self.ISUID and m & self.IEXEC:
			ret.append('s')
		elif m & self.IEXEC:
			ret.append('x')
		else:
			ret.append('-')

		m = self.getgroupmode()
		if m & self.IREAD:
			ret.append('r')
		else:
			ret.append('-')
		if m & self.IWRITE:
			ret.append('w')
		else:
			ret.append('-')
		if mode & self.ISGID and not m & self.IEXEC:
			ret.append('S')
		elif mode & self.ISGID and m & self.IEXEC:
			ret.append('s')
		elif m & self.IEXEC:
			ret.append('x')
		else:
			ret.append('-')

		m = self.getothermode()
		if m & self.IREAD:
			ret.append('r')
		else:
			ret.append('-')
		if m & self.IWRITE:
			ret.append('w')
		else:
			ret.append('-')
		if mode & self.ISVTX and not m & self.IEXEC:
			ret.append('T')
		elif mode & self.ISVTX and m & self.IEXEC:
			ret.append('t')
		elif m & self.IEXEC:
			ret.append('x')
		else:
			ret.append('-')

		return ''.join(ret)

class FFSInodeUFS2(dpkt.Packet):
	'''The on-disk structure for UFS2 inodes.'''

	__hdr__ = (
		('di_mode', 'H', 0),
		('di_nlink', 'h', -1),
		('di_uid', 'I', 0),
		('di_gid', 'I', 0),
		('di_blksize', 'I', 0),
		('di_size', 'Q', 0),
		('di_blocks', 'Q', 0),
		('di_atime', 'q', -1),
		('di_mtime', 'q', -1),
		('di_ctime', 'q', -1),
		('di_birthtime', 'q', -1),
		('di_mtimensec', 'i', -1),
		('di_atimensec', 'i', -1),
		('di_ctimensec', 'i', -1),
		('di_birthnsec', 'i', -1),
		('di_gen', 'i', -1),
		('di_kernflags', 'I', 0),
		('di_flags', 'I', 0),
		('di_extsize', 'i', -1),
		('di_extb', '%dq' % NXADDR, (-1, ) * NXADDR),
		('di_db', '%dq' % NDADDR, (-1, ) * NDADDR),
		('di_ib', '%dq' % NIADDR, (-1, ) * NIADDR),
		('di_spare', '3q', (-1, ) * 3),
		)
	__byte_order__ = endianness
assert FFSInodeUFS2.__hdr_len__ == 256

class FFSInodeUFS1(dpkt.Packet):
	'''The on-disk structure for UFS1 inodes.'''
	__hdr__ = (
		('di_mode', 'H', 0),
		('di_nlink', 'h', -1),
		('oldids', '2H', (0, ) * 2),
		('di_size', 'Q', 0),
		('di_atime', 'i', -1),
		('di_atimensec', 'i', -1),
		('di_mtime', 'i', -1),
		('di_mtimensec', 'i', -1),
		('di_ctime', 'i', -1),
		('di_ctimensec', 'i', -1),
		('di_db', '%di' % NDADDR, (-1, ) * NDADDR),
		('di_ib', '%di' % NIADDR, (-1, ) * NIADDR),
		('di_flags', 'I', 0),
		('di_blocks', 'i', -1),
		('di_gen', 'i', -1),
		('di_uid', 'I', 0),
		('di_gid', 'I', 0),
		('di_spare', '2i', (-1, ) * 2),
		)
	__byte_order__ = endianness
assert FFSInodeUFS1.__hdr_len__ == 128

class FFSInode(InodeConsts):
	'''Create an inode instance for a FFS filesystem.  Pass in the FFS
instance and the inode number.  It uses the FFS instance to decide
which on-disk structure to use.  Supports interation if the file
type is either regular or a directory.

There is an extra backup method where if you pass in an additional
string, it will use that string instead of fetching from disk.
The string must be the correct length for the format of this fs.
This is for debugging things like 8 byte offseted inode records.'''

	def __init__(self, fs, ino, *args):
		self.fs = fs
		self.ino = int(ino)
			
		if fs.sb.fs_magic == fs.sb.FS_UFS1_MAGIC:
			cls = FFSInodeUFS1
		else:
			cls = FFSInodeUFS2

		if len(args) > 1:
			raise TypeError, 'too many arguments'

		l = cls.__hdr_len__
		if len(args) == 0:
			data = fs.getblock(fs.sb.ino_to_fsba(ino))
			o = fs.sb.ino_to_fsbo(ino)
			data = data[o * l: (o + 1) * l]
		else:
			if len(args[0]) != l:
				raise TypeError, 'invalid length of string'
			data = args[0]
		self.dinode = cls(data)

	def __getattr__(self, name):
		'''This is used for pass through of the on-disk instance data.'''
		if name[:3] == 'di_':
			return getattr(self.dinode, name)

		raise NotImplementedError, name

	def maketime(self, sec, nsec):
		return sec + float(nsec) / 1000000000

	def atime(self):
		return self.maketime(self.di_atime, self.di_atimensec)

	def birthtime(self):
		return self.maketime(self.di_birthtime, self.di_birthtimensec)

	def ctime(self):
		return self.maketime(self.di_ctime, self.di_ctimensec)

	def mtime(self):
		return self.maketime(self.di_mtime, self.di_mtimensec)

	def restore(self, name):
		'''Restore the parameters on the inode.  This includes uid, gid,
mode, flags (if supported), and times.'''

		def ignoreoserror(fun):
			try:
				ret = fun()
			except OSError:
				return None
			return ret

		ignoreoserror(lambda: os.lchown(name, self.di_uid, self.di_gid))
		ignoreoserror(lambda: os.lchmod(name, self.getmode()))
		ignoreoserror(lambda: doflags(name, self.di_flags, self.getflags()))

		# XXX - no lutimes, so this could mess up other file's times
		try:
			ignoreoserror(lambda: os.utime(name, (self.atime(), self.birthtime())))
		except AttributeError:
			pass
		ignoreoserror(lambda: os.utime(name, (self.atime(), self.mtime())))

	def symlink(self):
		'''Return the value of the symlink.'''
		if self.di_size < self.fs.sb.fs_maxsymlinklen:
			args = [ '%s%d%s' % (self.dinode.__byte_order__, NDADDR + NIADDR, daddr_struct[self.fs.sb.getmagic()]) ] + self.di_db + self.di_ib
			return apply(struct.pack, args)[:self.di_size]
		return str(self)

	def alliblks(self, blk, n):
		'''An iterator that is used for handling indirect blocks.
You pass in the block number along with the number of still needed
indirections.'''

		if blk != 0:
			data = struct.unpack(self.fs.daddr_struct, self.fs.getblock(blk))
		else:
			data = itertools.repeat(0, self.fs.sb.fs_nindir)

		for i in data:
			if n:
				for j in self.alliblks(i, n - 1):
					yield j
			else:
				yield i

	def allblks(self):
		'''An iterator that will return all the possible blocks
referenced by the file.  This will continue past the file size.'''

		for i in self.di_db:
			yield i

		for p, i in enumerate(self.di_ib):
			for j in self.alliblks(i, p):
				yield j

	def fsblks(self):
		'''Returns all the blocks that are referenced by the
file.'''
		sz = self.di_size
		bsize = self.fs.sb.fs_bsize
		blks = self.allblks()
		while sz > 0:
			yield blks.next()
			sz -= bsize

	def datablocked(self):
		'''Iterator the returns all the data for this inode.  If
a block is sparse, it will be returned as the object at
self.fs.emptyblock, for quicker comparisions.'''

		# We have to keep a copy of the last block we received
		# before we yield it.  We need to do this so we know when
		# we've hit the last block, and only return the remainder
		# of that block.  If we didn't we'd return too much data
		# on files that aren't === self.di_size mod self.fs.sb.fs_bsize
		r = None
		emptyblock = self.fs.emptyblock
		getblock = self.fs.getblock
		for blk in self.fsblks():
			if r != None:
				yield r
			if blk != 0:
				r = getblock(blk)
			else:
				r = emptyblock

		if r == None:
			r = ''

		remain = self.di_size % self.fs.sb.fs_bsize
		if remain != 0:
			r = r[:remain]
		yield r

	def __str__(self):
		'''Returns the entire file as a string.  Make sure you
have enough memory to store it!'''

		return ''.join(self.datablocked())

	def __repr__(self):
		return '<FFSInode: %s, ino: %d, blocks: %d, %s nlink: %d, uid: %d, gid: %d, flags: %s, size: %d>' % (self.fs.sb.getmagic(), self.ino, self.di_blocks, self.lsfmt(), self.di_nlink, self.di_uid, self.di_gid, self.getflags(), self.di_size)

	def __iter__(self):
		if self.isdir():
			for i in self.datablocked():
				for d in DirBlock(i):
					yield d
		elif self.isreg():
			for i in self.datablocked():
				yield i
		else:
			raise NotImplementedError

class FFS:
	'''Master FFS class.  It takes a single argument which should be a
block like list of data (block.Block).  It will search for the
superblock in the standard locations.  The blocked argument must
start where the OS normally thinks that the filesystem starts.'''

	SBLOCK_FLOPPY = 0
	SBLOCK_UFS1 = 8192
	SBLOCK_UFS2 = 65536
	SBLOCK_PIGGY = 262144
	SBLOCK_SEARCH = [ 
		SBLOCK_FLOPPY,
		SBLOCK_UFS1,
		SBLOCK_UFS2,
		SBLOCK_PIGGY,
		]

	def __init__(self, f):
		self.f = f
		foundsb = False
		for i in self.SBLOCK_SEARCH:
			sb = getsboff(f, i)
			if sb.isvalid():
				foundsb = True
				break

		if not foundsb:
			raise ValueError, "no valid superblock found"

		if f.blocksize != DEV_BSIZE:
			raise ValueError, 'f does not have compatible blocksize'

		self.sb = sb

		cg = []
		for i in range(sb.fs_ncg):
			tcg = FFSCylinderGroup(f.getlen(sb.fsbtodb(sb.cgtod(i)), sb.fs_cgsize))
			# do we really want to enfoce this?
			if not tcg.isvalid():
				raise ValueError, "cylindergroup %d of %d not valid" % (i, sb.fs_ncg)
			cg.append(tcg)

		self.cg = cg
		self.inodes = weakref.WeakValueDictionary()
		self.daddr_struct = ''.join((endianness, str(self.sb.fs_nindir),
		    daddr_struct[self.sb.getmagic()]))
		assert struct.calcsize(self.daddr_struct) == self.sb.fs_bsize
		self.emptyblock = "\x00" * self.sb.fs_bsize

	def getblock(self, blk):
		return self.f.getlen(self.sb.fsbtodb(blk), self.sb.fs_bsize)

	def getfrag(self, frag):
		return self.f.getlen(self.sb.fsbtodb(frag), sb.fs_fsize)

	def getinode(self, ino):
		try:
			return self.inodes[ino]
		except KeyError:
			ret = FFSInode(self, ino)
			self.inodes[ino] = ret
			return ret

	def __getitem__(self, key):
		try:
			return self.getinode(int(key))
		except ValueError:
			# key is a path
			pass

	def __iter__(self):
		for c in range(self.sb.fs_ncg):
			if self.sb.fs_magic == self.sb.FS_UFS1_MAGIC:
				cnt = self.sb.fs_ipg
			else:
				cnt = self.cg[c].cg_initediblk
			for i in range(cnt):
				yield self[c * self.sb.fs_ipg + i]

	alldirs = lambda self: ifilter(lambda x: x.isdir(), self)
	allregs = lambda self: ifilter(lambda x: x.isreg(), self)

def getoffsetof(cls, member, l = 8192):
	'''Find the offset of a member of a dpkt class.'''
	dummydata = ''.join([struct.pack('H', i * 2) for i in range(l / 2)])
	m = getattr(cls(dummydata), member)
	return struct.unpack('H', struct.pack('<i', m)[:2])[0]

sbmagicoff = getoffsetof(FFSSuperBlock, 'fs_magic')

def findsuperblocks(f):
	'''Return a list of blocks that start a valid superblock.'''

	magicsuboff = sbmagicoff % f.blocksize
	magicsubdiv = sbmagicoff / f.blocksize
	FS_MAGIC = FFSSuperBlock.FS_MAGIC
	ffsmagics = FS_MAGIC.keys()
	magicstruct = FFSSuperBlock.__byte_order__ + 'i'
	magiclen = struct.calcsize(magicstruct)
	ret = []
	for b, i in enumerate(f):
		cv = struct.unpack(magicstruct, i[magicsuboff:magicsuboff + magiclen])[0]
		if cv in ffsmagics:
			ret.append(b - magicsubdiv)

	return ret

def findstartfs(f, blk):
	'''Returns the block that is the start of the file system for the
superblock located at blk.  If blk is a backup superblock, and
there is a valid cylinder group following, it will locate the
master superblock, and return the start of the fs.'''

	oblk = blk
	osb = sb = getsbblk(f, blk)
	assert sb.isvalid()
	assert f.blocksize <= sb.fs_fsize
	assert sb.check()

	#Now that we have a valid SuperBlock, try to find if we are the first
	#(fs_sblockloc) or the second (fs_sblkno)

	# XXX - Just use the cylindar method if we are the second.  Much less
	# code.
	second = blk + (sb.fs_sblkno * sb.fs_fsize - sb.fs_sblockloc) / f.blocksize
	first =	blk + (sb.fs_sblockloc - sb.fs_sblkno * sb.fs_fsize) / f.blocksize

	if not getsbblk(f, second).isvalid():
		# Ok, get the first one, it better be valid
		sb = getsbblk(f, first)
		blk = first

	if not sb.isvalid():
		# now check to see if there is a cylindar group after the
		# passed in superblock, if there is, figure out where the
		# start should be and check for that.
		sb = osb
		blk = oblk
		cg = FFSCylinderGroup(f.getlen(oblk + (sb.fs_cblkno - sb.fs_sblkno) * sb.fs_fsize / f.blocksize, sb.fs_bsize))
		if cg.isvalid():
			first = blk - (sb.fs_fpg * cg.cg_cgx + sb.fs_sblkno) * (sb.fs_fsize / f.blocksize) + sb.fs_sblockloc / f.blocksize
			sb = getsbblk(f, first)
			if sb.isvalid():
				blk = first
			else:
				raise ValueError, "valid cylinder group, but no valid superblock"
		else:
			raise ValueError, "both superblocks not found"
	assert sb.check()

	return blk - sb.fs_sblockloc / f.blocksize

def callorprint(cop, *args):
	'''If cop is callable, call it and return, otherwise apply args to
cop using the `%' operator.'''

	if callable(cop):
		return apply(cop, args)
	else:
		return cop % args

def printsb(f, off):
	'''Print out a superblock is a human friendly/readable manner.'''
	sb = getsbblk(f, off)

	print 'Magic: %s' % sb.getmagic()
	oneline = (
		('fsmnt', 'fs_fsmnt', '%s'),
		('volname', 'fs_volname', '%s'),
		('time', 'fs_time', '%18d'),
		('maxfilesize', 'fs_maxfilesize', '%18d'),
		)
	maxheader = reduce(max, map(lambda x: len(x[0]), oneline))
	for i in oneline:
		print ('%*s: ' + i[2]) % (maxheader, i[0], getattr(sb, i[1]))

	general = (
		('sblkno', 'fs_sblkno', '%*d'),
		('cblkno', 'fs_cblkno', '%*d'),
		('iblkno', 'fs_iblkno', '%*d'),
		('dblkno', 'fs_dblkno', '%*d'),
		('ncg', 'fs_ncg', '%*d'),
		('bsize', 'fs_bsize', '%*d'),
		('fsize', 'fs_fsize', '%*d'),
		('frag', 'fs_frag', '%*d'),
		('minfree', 'fs_minfree', '%*d'),
		('bmask', 'fs_bmask', '%#*x'),
		('fmask', 'fs_fmask', '%#*x'),
		('bshift', 'fs_bshift', '%*d'),
		('fshift', 'fs_fshift', '%*d'),
		('maxcontig', 'fs_maxcontig', '%*d'),
		('maxbpg', 'fs_maxbpg', '%*d'),
		('fragshift', 'fs_fragshift', '%*d'),
		('fsbtodb', 'fs_fsbtodb', '%*d'),
		('sbsize', 'fs_sbsize', '%*d'),
		('nindir', 'fs_nindir', '%*d'),
		('inopb', 'fs_inopb', '%*d'),
		('optim', 'fs_optim', '%*d'),
		('cssize', 'fs_cssize', '%*d'),
		('cgsize', 'fs_cgsize', '%*d'),
		('ipg', 'fs_ipg', '%*d'),
		('fpg', 'fs_fpg', '%*d'),
		('fmod', 'fs_fmod', '%*d'),
		('clean', 'fs_clean', '%*d'),
		('ronly', 'fs_ronly', '%*d'),
		('swuid', 'fs_swuid', '%*d'),
		('cgrotor', 'fs_cgrotor', '%*d'),
		('maxbsize', 'fs_maxbsize', '%*d'),
		('sblockloc', 'fs_sblockloc', '%*d'),
		('cs_ndir', 'fs_cstotal_cs_ndir', '%*d'),
		('cs_nbfree', 'fs_cstotal_cs_nbfree', '%*d'),
		('cs_nifree', 'fs_cstotal_cs_nifree', '%*d'),
		('cs_nffree', 'fs_cstotal_cs_nffree', '%*d'),
		('cs_numclusters', 'fs_cstotal_cs_numclusters', '%*d'),
		('size', 'fs_size', '%*d'),
		('dsize', 'fs_dsize', '%*d'),
		('csaddr', 'fs_csaddr', '%*d'),
		('pendingblocks', 'fs_pendingblocks', '%*d'),
		('pendinginodes', 'fs_pendinginodes', '%*d'),
		('avgfilesize', 'fs_avgfilesize', '%*d'),
		('avgfpdir', 'fs_avgfpdir', '%*d'),
		('save_cgsize', 'fs_save_cgsize', '%*d'),
		('flags', 'fs_flags', '%*d'),
		('contigsumsize', 'fs_contigsumsize', '%*d'),
		('maxsymlinklen', 'fs_maxsymlinklen', '%*d'),
		('qbmask', 'fs_qbmask', '%*x'),
		('qfmask', 'fs_qfmask', '%*x'),
		('state', 'fs_state', '%*d'),
		)
	oldstuff = (
		('cgoffset', 'fs_old_cgoffset', '%d'),
		('cgmask', 'fs_old_cgmask', '%d'),
		('time', 'fs_old_time', '%d'),
		('size', 'fs_old_size', '%d'),
		('dsize', 'fs_old_dsize', '%d'),
		('rotdelay', 'fs_old_rotdelay', '%d'),
		('rps', 'fs_old_rps', '%d'),
		('nspf', 'fs_old_nspf', '%d'),
		('npsect', 'fs_old_npsect', '%d'),
		('interleave', 'fs_old_interleave', '%d'),
		('trackskew', 'fs_old_trackskew', '%d'),
		('csaddr', 'fs_old_csaddr', '%d'),
		('nsect', 'fs_old_nsect', '%d'),
		('spc', 'fs_old_spc', '%d'),
		('ncyl', 'fs_old_ncyl', '%d'),
		('cpg', 'fs_old_cpg', '%d'),
		('flags', 'fs_old_flags', '%d'),
		('cpc', 'fs_old_cpc', '%d'),
		('inodefmt', 'fs_old_inodefmt', '%d'),
		('postblformat', 'fs_old_postblformat', '%d'),
		('nrpos', 'fs_old_nrpos', '%d'),
		)
	itemstoprint = general
	maxheader = reduce(max, map(lambda x: len(x[0]), itemstoprint))
	maxdlen = reduce(max, map(lambda i, sb = sb: len(callorprint(i[2], 0, getattr(sb, i[1]))), itemstoprint))
	# ncols * (maxdlen + maxheader + 2) + ncols - 1 = 80
	# ncols * (maxdlen + maxheader + 2) + ncols = 81
	# ncols * (maxdlen + maxheader + 2 + 1) = 81
	ncols = 81 / (maxdlen + maxheader + 3)
	data = map(lambda i, mh = maxheader, md = maxdlen, sb = sb: '%*s: ' % (mh, i[0])  + callorprint(i[2], md, getattr(sb, i[1])), itemstoprint)
	while data:
		print ' '.join(data[:ncols])
		del data[:ncols]

def printcg(fs, cgnum):
	'''Print out a cyliner group is a human friendly/readable manner.'''

	general = (
		('cgx', 'cg_cgx', '%*d'),
		('ndblk', 'cg_ndblk', '%*d'),
		('cs_ndir', 'cg_cs_ndir', '%*d'),
		('cs_nbfree', 'cg_cs_nbfree', '%*d'),
		('cs_nifree', 'cg_cs_nifree', '%*d'),
		('cs_nffree', 'cg_cs_nffree', '%*d'),
		('rotor', 'cg_rotor', '%*d'),
		('frotor', 'cg_frotor', '%*d'),
		('irotor', 'cg_irotor', '%*d'),
		('iusedoff', 'cg_iusedoff', '%*d'),
		('freeoff', 'cg_freeoff', '%*d'),
		('nextfreeoff', 'cg_nextfreeoff', '%*d'),
		('clustersumoff', 'cg_clustersumoff', '%*d'),
		('clusteroff', 'cg_clusteroff', '%*d'),
		('nclusterblks', 'cg_nclusterblks', '%*d'),
		('niblk', 'cg_niblk', '%*d'),
		('initediblk', 'cg_initediblk', '%*d'),
		)

	cg = fs.cg[cgnum]
	itemstoprint = general
	maxheader = reduce(max, map(lambda x: len(x[0]), itemstoprint))
	maxdlen = reduce(max, map(lambda i, cg = cg: len(callorprint(i[2], 0, getattr(cg, i[1]))), itemstoprint))
	# ncols * (maxdlen + maxheader + 2) + ncols - 1 = 80
	# ncols * (maxdlen + maxheader + 2) + ncols = 81
	# ncols * (maxdlen + maxheader + 2 + 1) = 81
	ncols = 81 / (maxdlen + maxheader + 3)
	data = map(lambda i, mh = maxheader, md = maxdlen, cg = cg: '%*s: ' % (mh, i[0])  + callorprint(i[2], md, getattr(cg, i[1])), itemstoprint)
	while data:
		print ' '.join(data[:ncols])
		del data[:ncols]

def dumpinode(inode, name="%(ino)d"):
	f = open(name % inode, "w")
	filter(f.write, inode.datablocked())

def restoreinum(fs, ino, name, reexclude):
	i = fs[ino]
	if i.isdir():
		os.mkdir(name)
		for j in i:
			if j.name in ('.', '..') or reexclude.match(j.name):
				continue
			restoreinum(fs, j.ino, os.path.join(name, j.name), reexclude)
	elif i.isreg():
		dumpinode(i, name)
	elif i.islnk():
		os.symlink(i.symlink(), name)
	else:
		print 'Cannot restore inode %d as %s, unkown inode format.' % (i.ino, name)
		return

	i.restore(name)

def usage():
	sys.stderr.write("Usage: prog -a | -c <cgnum> | -d | -f | -i <inum> | -l <inum> file\n")
	sys.stderr.write("Usage: prog -p | -s -o blk file\n")
	sys.stderr.write("Usage: prog [ -x <exregex> ] -r <inum> -n <name> file\n")
	sys.stderr.write("Args for all: [ -b <beginblk> ] | [ -z <filesize> ]\n")
	sys.stderr.write('''Description:
	-a	Dump all regular inodes as numbered files in cwd
	-c	Print out details of cylinder group
	-d	Dump out all directories, and their entries
	-f	Find all super blocks with a valid magic
	-i	Dump inode's data to stdout
	-l	Print out inode's contains
	-n	Name of inode to restore to
	-p	Print out superblock at offset -o
	-r	Inode to start recursive restore at (requries -n)
	-s	Print out fs start offset using super block at -o
	-x	exclude files named by the regular expresion from restore
''')

def main():
	try:
		opts, args = getopt.getopt(sys.argv[1:], "ab:c:dfhi:l:n:o:pr:sx:z:")
	except getopt.GetoptError:
		# print help information and exit:
		usage()
		sys.exit(2)

	doalldump = False
	beginblk = 0
	docgnum = None
	dodirdump = False
	dosuper = False
	doinumdump = None
	doinumdata = None
	doinumrestore = None
	restoretoname = None
	sboff = None
	doprintsb = False
	fsstart = False
	filesize = None
	namereexclude = None

	needsboff = ('doprintsb', 'fsstart', )
	needfs = ('doinumdump', 'docgnum', 'dodirdump', 'doinumdata', 'doalldump', 'doinumrestore', )

	for o, a in opts:
		if o == "-a":
			doalldump = True
		elif o == "-b":
			beginblk = int(a)
		elif o == '-c':
			docgnum = int(a)
		elif o == '-d':
			dodirdump = True
		elif o == '-f':
			dosuper = True
		elif o == "-h":
			usage()
			sys.exit()
		elif o == '-i':
			doinumdump = int(a)
		elif o == '-l':
			doinumdata = int(a)
		elif o == '-n':
			restoretoname = a
		elif o == '-p':
			doprintsb = True
		elif o == '-o':
			sboff = int(a)
		elif o == '-r':
			doinumrestore = int(a)
		elif o == '-s':
			fsstart = True
		elif o == '-x':
			namereexclude = re.compile(a)
		elif o == '-z':
			filesize = int(a)

	if filter(None, map(lambda x, y=locals(): y[x], needsboff)) and sboff == None:
		usage()
		sys.exit(2)

	rargs = (doinumrestore, restoretoname)
	if filter(lambda x: x != None, rargs) and filter(lambda x: x == None, rargs):
		sys.stderr.write('Must provide both -r and -n.\n')
		usage()
		sys.exit(2)

	if len(args) != 1:
		usage()
		sys.exit(2)

	f = block.Block(open(args[0]), offset=beginblk, length=filesize)

	# If we need fs, create it
	if filter(lambda x: x != None and x != False, map(lambda x, y=locals(): y[x], needfs)):
		fs = FFS(f)

	if doalldump:
		filter(dumpinode, fs.allregs())
	elif docgnum != None:
		printcg(fs, docgnum)
	elif dodirdump:
		#filter(lambda x: sys.stdout.write('%s\n' % x), imap(repr, ifilter(lambda x: x.isdir(), fs)))
		sys.stdout.writelines(imap(lambda x: '%s: %s\n' % (repr(x), ', '.join(imap(repr, x))), fs.alldirs()))
		#filter(lambda x: sys.stdout.write('%s\n' % x), map(lambda x: map(repr, x), filter(lambda x: x.isdir(), fs)))
	elif dosuper:
		print ' '.join(map(str, findsuperblocks(f)))
	elif doinumdump != None:
		filter(sys.stdout.write, fs[doinumdump].datablocked())
	elif doinumdata != None:
		print repr(fs[doinumdata])
	elif doinumrestore:
		restoreinum(fs, doinumrestore, os.path.join(os.getcwd(), restoretoname), namereexclude)
	elif doprintsb:
		printsb(f, sboff)
	elif fsstart != None:
		print findstartfs(f, sboff)

if __name__ == '__main__':
	main()
